/* $Header$ */

/* Copyright (c) 1999 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  os_stdio.cpp - OS implementation for standard I/O
Function
  This is a simple stdio-based implementation of the OS display functions.
  This is a portable implementation (at least, it's portable to any
  platform where stdio can be used).

  The main purpose of this implementation is to make it easy to build
  command-line versions of tools that incorporate osifc-layer functions
  without having to drag in the full-blown OS-specific implementation.
  For example, test tools might find this useful.
Notes
  
Modified
  09/04/99 MJRoberts  - Creation
*/

#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "os.h"
#include "t3std.h"

/* 
 *   status-line display mode 
 */
static int S_status_mode = 0;

/* 
 *   initialize 
 */
int os_init(int *, char **, const char *, char *, int)
{
    return 0;
}

/*
 *   uninitialize 
 */
void os_uninit()
{
}

/*
 *   flush output 
 */
void os_flush()
{
}

/*
 *   get highlighting character - indicate that highlighting is to be
 *   ignored 
 */
int os_hilite(int /*mode*/)
{
    /* no highlighting available - return zero */
    return 0;
}

/*
 *   set the title 
 */
void os_set_title(const char *)
{
    /* do nothing in this implementation */
}

/*
 *   display text 
 */
void os_printf(const char *fmt, ...)
{
    va_list argptr;

    /* if we're not in main text mode, ignore the text */
    if (S_status_mode != 0)
        return;

    /* pass the text through to stdio */
    va_start(argptr, fmt);
    vprintf(fmt, argptr);
    va_end(argptr);
}

/*
 *   show the mouse cursor - do nothing in this implementation
 */
void os_moushow()
{
    /* do nothing in this implementation */
}

/* 
 *   hide the mouse cursor - do nothing in this implementation 
 */
void os_mouhide()
{
    /* do nothing in this implementation */
}

/*
 *   get an event 
 */
int os_get_event(unsigned long timeout, int use_timeout,
                 os_event_info_t *info)
{
    /* we can't handle timeouts */
    if (use_timeout)
        return OS_EVT_NOTIMEOUT;

    /* get a key */
    info->key[0] = os_getc_raw();
    if (info->key[0] == 0)
        info->key[1] = os_getc_raw();

    /* return the keyboard event */
    return OS_EVT_KEY;
}

/*
 *   clear the screen 
 */
void oscls()
{
    /* 
     *   print a ^L to clear the screen, if such sequences are interpreted
     *   by the display driver or terminal 
     */
    printf("\014\n");
}

/*
 *   read from the keyboard 
 */
unsigned char *os_gets(unsigned char *buf, size_t)
{
    return (unsigned char *)gets((char *)buf);
}

/*
 *   prompt for information through a dialog
 */
int os_input_dialog(int /*icon_id*/,
                    const char *prompt, int standard_button_set,
                    const char **buttons, int button_count,
                    int /*default_index*/, int /*cancel_index*/)
{
    /* keep going until we get a valid response */
    for (;;)
    {
        int i;
        char buf[256];
        const char *p;
        const char *cur;
        char *resp;
        int match_cnt;
        int last_found;
        const static struct
        {
            const char *buttons[3];
            int button_count;
        } std_btns[] =
        {
            { { "&OK" },                    1 },
            { { "&OK", "&Cancel" },         2 },
            { { "&Yes", "&No" },            2 },
            { { "&Yes", "&No", "&Cancel" }, 3 }
        };

        /* 
         *   if we have a standard button set selected, get our button
         *   labels 
         */
        switch(standard_button_set)
        {
        case 0:
            /* use the explicit buttons provided */
            break;

        case OS_INDLG_OK:
            i = 0;

        use_std_btns:
            /* use the selected standard button set */
            buttons = (const char **)std_btns[i].buttons;
            button_count = std_btns[i].button_count;
            break;

        case OS_INDLG_OKCANCEL:
            i = 1;
            goto use_std_btns;

        case OS_INDLG_YESNO:
            i = 2;
            goto use_std_btns;

        case OS_INDLG_YESNOCANCEL:
            i = 3;
            goto use_std_btns;

        default:
            /* 
             *   we don't recognize other standard button sets - return an
             *   error 
             */
            return 0;
        }

        /* 
         *   if there are no buttons defined, they'll never be able to
         *   respond, so we'd just loop forever - rather than let that
         *   happen, return failure 
         */
        if (button_count == 0)
            return 0;

        /* display a newline and the prompt string */
        os_printf("\n%s ", prompt);

        /* display the response */
        for (i = 0 ; i < button_count ; ++i)
        {
            /* 
             *   display a slash to separate responses, if this isn't the
             *   first one 
             */
            if (i != 0)
                os_printf("/");

            /* get the current button */
            cur = buttons[i];

            /* 
             *   Look for a "&" in the response string.  If we find it,
             *   remove the "&" and enclose the shortcut key in parens.  
             */
            for (p = cur ; *p != '&' && *p != '\0' ; ++p) ;

            /* if we found the "&", put the next character in parens */
            if (*p == '&')
            {
                /* reformat the response string */
                sprintf(buf, "%.*s(%c)%s", (int)(p - cur), cur, *(p+1), p+2);

                /* show it */
                os_printf("%s", buf);
            }
            else
            {
                /* no '&' - just display the response string as-is */
                os_printf("%s", cur);
            }
        }

        /* read the response */
        os_printf(" > ");
        os_gets((unsigned char *)buf, sizeof(buf));

        /* skip any leading spaces in the reply */
        for (resp = buf ; isspace(*resp) ; ++resp) ;

        /* if it's one character, check it against the shortcut keys */
        if (strlen(resp) == 1)
        {
            /* scan the responses */
            for (i = 0 ; i < button_count ; ++i)
            {
                /* look for a '&' in this button */
                for (p = buttons[i] ; *p != '&' && *p != '\0' ; ++p) ;

                /* if we found the '&', check the shortcut */
                if (*p == '&' && toupper(*(p+1)) == toupper(*resp))
                {
                    /* 
                     *   this is the one - return the current index
                     *   (bumping it by one to get a 1-based value) 
                     */
                    return i + 1;
                }
            }
        }

        /* 
         *   Either it's not a one-character reply, or it didn't match a
         *   short-cut - check it against the leading substrings of the
         *   responses.  If it matches exactly one of the responses in its
         *   leading substring, use that response.  
         */
        for (i = 0, match_cnt = 0 ; i < button_count ; ++i)
        {
            const char *p1;
            const char *p2;

            /* 
             *   compare this response to the user's response; skip any
             *   '&' in the button label 
             */
            for (p1 = resp, p2 = buttons[i] ; *p1 != '\0' && *p2 != '\0' ;
                 ++p1, ++p2)
            {
                /* if this is a '&' in the button label, skip it */
                if (*p2 == '&')
                    ++p2;

                /* if these characters don't match, it's no match */
                if (toupper(*p1) != toupper(*p2))
                    break;
            }

            /* 
             *   if we reached the end of the user's response, we have a
             *   match in the leading substring - count it and remember
             *   this as the last one, but keep looking, since we need to
             *   make sure we don't have any other matches 
             */
            if (*p1 == '\0')
            {
                ++match_cnt;
                last_found = i;
            }
        }

        /* 
         *   if we found exactly one match, return it (adjusting to a
         *   1-based index); if we found more or less than one match, it's
         *   not a valid response, so start over with a new prompt 
         */
        if (match_cnt == 1)
            return last_found + 1;
    }
}

/*
 *   ask for a file 
 */
int os_askfile(const char *prompt, char *reply, int replen,
               int /*dialog_type*/, os_filetype_t /*file_type*/)
{
    /* show the prompt */
    os_printf("%s >", prompt);

    /* ask for the filename */
    os_gets((unsigned char *)reply, replen);

    /* 
     *   if they entered an empty line, return "cancel"; otherwise, return
     *   success 
     */
    return (reply[0] == '\0' ? OS_AFE_CANCEL : OS_AFE_SUCCESS);
}

/*
 *   get system information 
 */
int os_get_sysinfo(int code, void *param, long *result)
{
    switch(code)
    {
    case SYSINFO_HTML:
    case SYSINFO_JPEG:
    case SYSINFO_PNG:
    case SYSINFO_WAV:
    case SYSINFO_MIDI:
    case SYSINFO_WAV_MIDI_OVL:
    case SYSINFO_WAV_OVL:
    case SYSINFO_PREF_IMAGES:
    case SYSINFO_PREF_SOUNDS:
    case SYSINFO_PREF_MUSIC:
    case SYSINFO_PREF_LINKS:
    case SYSINFO_MPEG:
    case SYSINFO_MPEG1:
    case SYSINFO_MPEG2:
    case SYSINFO_MPEG3:
        /* 
         *   we don't support any of these features - set the result to 0
         *   to indicate this 
         */
        *result = 0;

        /* return true to indicate that we recognized the code */
        return TRUE;

    default:
        /* we don't recognize other codes */
        return FALSE;
    }
}

/*
 *   set status mode
 */
void os_status(int mode)
{
    /* remember the new mode */
    S_status_mode = mode;
}

/*
 *   display a string in the right half of the status line 
 */
void os_strsc(const char *)
{
    /* ignore it - we don't have a status line in this UI */
}

void os_plain()
{
    /* 
     *   this implementation is always in plain mode, so there's nothing
     *   we need to do here 
     */
}

