/* $Header$ */

/* Copyright (c) 1999 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  vmconsol.h - TADS 3 console input reader and output formatter
Function
  Provides console input and output for the TADS 3 built-in function set
  for the T3 VM.

  T3 uses the UTF-8 character set to represent character strings.  The OS
  functions use the local character set.  We perform the mapping between
  UTF-8 and the local character set within this module, so that OS routines
  see local characters only, not UTF-8.

  This code is based on the TADS 2 output formatter, but has been
  substantially reworked for C++, Unicode, and the slightly different
  TADS 3 formatting model.
Notes
  
Modified
  09/04/99 MJRoberts  - Creation
*/

#ifndef VMCONSOL_H
#define VMCONSOL_H

#include <string.h>
#include <stdlib.h>
#include <wchar.h>

#include "os.h"
#include "vmglob.h"


/* ------------------------------------------------------------------------ */
/*
 *   MORE mode - this determines whether we handle MORE prompts in the
 *   output formatter, or let the underlying OS display code handle MORE
 *   prompting.
 *   
 *   The OS header previously included should #define USE_MORE if we
 *   should handle the MORE prompt in the output formatter, or leave this
 *   undefined if not.
 *   
 *   Note that we specifically turn off HTML mode if we are being compiled
 *   for use with an HTML renderer.  The HTML output code, which to us
 *   looks like an ordinary OS layer (via the standard interfaces -
 *   os_printf and so on), handles MORE prompting itself.  
 */

/* in HTML mode, don't use MORE mode */
#ifdef USE_HTML
# ifdef USE_MORE
#  undef USE_MORE
# endif
#endif


/* ------------------------------------------------------------------------ */
/*
 *   Newline type codes.  These specify how we are to perform line
 *   breaking after writing out text.
 */
enum vm_nl_type
{
    /* 
     *   no line separation at all - write out this text and subsequent
     *   text as part of the same line with no separators 
     */
    VM_NL_NONE,

    /* break the line at the end of this text and start a newline */
    VM_NL_NEWLINE,

    /* OS line separation - add a space after the text */
    VM_NL_OSNEWLINE,

    /* 
     *   flushing internal buffers only: no line separation, and do not
     *   flush to underlying OS level yet 
     */
    VM_NL_NONE_INTERNAL
};


/* ------------------------------------------------------------------------ */
/*
 *   Console.  A console corresponds to device that shows information to
 *   and reads text from the user.  On a text system, the console is
 *   simply the terminal.  On a graphical system, the console is usually
 *   an application window.  
 */
class CVmConsole
{
public:
    CVmConsole();
    ~CVmConsole();

    /* write out a null-terminated UTF-8 string */
    int format_text(VMG_ const char *p)
    {
        /* get its length and write it out */
        return format_text(vmg_ p, strlen(p));
    }

    /* write out a UTF-8 string of a given byte length */
    int format_text(VMG_ const char *p, size_t len);

    /* display a blank line */
    void write_blank_line(VMG0_);

    /* set the whitespace mode (returns the old whitespace mode) */
    int set_obey_whitespace(int f);

    /* set the caps flag - capitalize the next character */
    void caps();

    /* set the nocaps flag - make the next letter miniscule */
    void nocaps();

    /* flush the output with the given newline type */
    void flush(VMG_ vm_nl_type nl);

    /* flush the output, ending the current line and starting a new line */
    void flush(VMG0_) { flush(vmg_ VM_NL_NEWLINE); }

    /* 
     *   Open a new log file.  Closes any previous log file.  If the file
     *   already exists, we'll overwrite it with the new log information,
     *   otherwise we'll create a new file.  Returns zero on success,
     *   non-zero on failure.  
     */
    int open_log_file(const char *fname);

    /* 
     *   close any existing log file - returns zero on success, non-zero
     *   on failure 
     */
    int close_log_file();

    /*
     *   Open a new command log file.  We'll log commands (and only
     *   commands) to the command log file.  Returns zero on success,
     *   non-zero on failure.  
     */
    int open_command_log(const char *fname);

    /* close the command log file, if there is one */
    int close_command_log();

    /*
     *   Set the current MORE mode.  Returns the old state.  The state is
     *   true if we show MORE prompts, false if not.  The state will be
     *   false if the underlying OS display layer handles prompting, so a
     *   return of false doesn't necessarily mean that MORE prompts are
     *   never shown, but merely that we don't handle MORE prompts in the
     *   output formatter itself.  
     */
    int set_more_state(int state)
    {
        int old_state;

        /* remember the old state */
        old_state = G_os_moremode;

        /* set the new mode */
        G_os_moremode = state;

        /* return the previous state */
        return old_state;
    }

    /* determine if we're in MORE mode */
    int is_more_mode() const { return G_os_moremode; }

    /* show the MORE prompt and wait for the user to acknowledge it */
    void show_more_prompt(VMG0_);

    /* get/set the line width of the display device */
    int get_line_width() const { return G_os_linewidth; }
    void set_line_width(int cols) { G_os_linewidth = cols; }

    /* get/set the page length of the display device */
    int get_page_length() const { return G_os_pagelength; }
    void set_page_length(int lines) { G_os_pagelength = lines; }

    /* get/set the double-space flag (for periods and other punctuation) */
    int get_doublespace() const { return doublespace_; }
    void set_doublespace(int f) { doublespace_ = f; }

    /* reset the MORE prompt line count */
    void reset_line_count();

    /* determ ine if we're in HTML mode */
    int is_html_mode();

    /*
     *   Start/end watchpoint evaluation.  During watchpoint evaluation,
     *   we suppress all output to the main text window, to avoid
     *   cluttering the main window with output generated as a side effect
     *   of evaluating a watchpoint expression.  
     */
    void set_wx_flag(int flag) { outwxflag_ = flag; }

    /* 
     *   check to see if we're reading from a script input file - returns
     *   true if so, false if reading from the user (via the keyboard or
     *   other input device) 
     */
    int is_reading_script() const { return (script_fp_ != 0); }

    /* 
     *   check to see if we're reading quietly from a script - if we're
     *   reading from a script, and this flag is true, we suppress all
     *   output
     */
    int is_quiet_script() const { return quiet_script_; }

    /* 
     *   Open a script file.  If 'quiet' is true, no output is displayed
     *   while the script is being processed.  If 'script_more_mode' is
     *   true, MORE mode is in effect while processing the script,
     *   otherwise MORE mode is turned off while processing the script (to
     *   leave things as they are, simply pass in is_more_mode() for this
     *   argument).  
     */
    void open_script_file(const char *fname, int quiet, int script_more_mode);

    /* 
     *   Close the script file.  Returns the original MORE mode that was
     *   in effect before the script file was opened; this MORE mode
     *   should be restored.  
     */
    int close_script_file();

    /* 
     *   Read a line of text from the keyboard.  Fills in the buffer with
     *   a null-terminated UTF-8 string.  Returns zero on success,
     *   non-zero on end-of-file reading the console (which usually
     *   indicates that the user has closed the application, so we're in
     *   the process of terminating; it might also indicate that the
     *   user's terminal has been detached, in which case we also probably
     *   can't do much except terminate).  
     */
    int read_line(VMG_ char *buf, size_t buflen);

    /*
     *   Display a file dialog.  This routine works exactly the same way
     *   as os_askfile(), but is implemented here to allow for a formatted
     *   text interface on systems where no dialog is available.  
     */
    int askfile(VMG_ const char *prompt, size_t prompt_len,
                char *reply, size_t replen,
                int dialog_type, os_filetype_t file_type);

    /*
     *   Display a system dialog.  This routine works exactly the same way
     *   as os_input_dialog(), but is implemented here to allow for a
     *   formatted text interface on systems where no dialog is available.
     */
    int input_dialog(VMG_ int icon_id, const char *prompt,
                     int standard_button_set,
                     const char **buttons, int button_count,
                     int default_index, int cancel_index);

protected:
    /* read a line from the script file */
    int read_line_from_script(char *buf, size_t buflen);

    /* our display stream - this stream shows up on the display */
    class CVmFormatterDisp *disp_str_;

    /* our log stream - this stream is written to the log file, if any */
    class CVmFormatterLog *log_str_;
    
    /*
     *   Flag: display two spaces after a period-like punctuation mark.
     *   This should be true if the output should have two spaces after a
     *   period, question mark, or exclamation point, false for a single
     *   space.  This should generally be true for fixed-width fonts,
     *   false for proportional fonts, although some users might prefer to
     *   use single-spacing even for fixed-width fonts.  
     */
    unsigned int doublespace_ : 1;

    /* watchpoint mode flag */
    unsigned int outwxflag_ : 1;

    /* 
     *   input script file - this will be null if we're reading directly
     *   from the keyboard, and will be a file pointer if we're reading
     *   input from a script file 
     */
    osfildef *script_fp_;

    /* flag: reading quietly from script input file */
    unsigned int quiet_script_ : 1;

    /* flag: MORE mode that was in effect before the current script file */
    unsigned int pre_script_more_mode_ : 1;

    /* command log file, if there is one */
    osfildef *command_fp_;

    /* name of the command log file */
    char command_fname_[OSFNMAX];
};


/* ------------------------------------------------------------------------ */
/*
 *   constants 
 */

/* 
 *   size of our HTML title buffer - this is used for accumulating the
 *   text enclosed in a <TITLE> tag 
 */
const size_t HTML_TITLE_BUF_SIZE = 256;

/*
 *   QSPACE is the special character for a quoted space (internally, the
 *   sequence "\ " (backslash-space) is converted to QSPACE).  It must not
 *   be any printable character.  The value here may need to be changed in
 *   the extremely unlikely event that TADS is ever ported to an EBCDIC
 *   machine.
 */
#define QSPACE 26

/*
 *   QTAB is a special hard tab character indicator.  We use this when we
 *   need to generate a hard tab to send to the underlying output layer
 *   (in particular, we use this to send hard tabs to the HTML formatter
 *   when we're in HTML mode).  
 */
#define QTAB 25


/*
 *   HTML lexical analysis modes 
 */
enum vmconsole_html_mode
{
    HTML_MODE_NORMAL,                      /* normal text, not in a tag */
    HTML_MODE_TAG,                                  /* parsing inside a tag */
    HTML_MODE_SQUOTE,                 /* in a single-quoted string in a tag */
    HTML_MODE_DQUOTE                  /* in a double-quoted string in a tag */
};

/*
 *   HTML parsing mode flag for <BR> tags.  We defer these until we've
 *   read the full tag in order to obey an HEIGHT attribute we find.  When
 *   we encounter a <BR>, we figure out whether we think we'll need a
 *   flush or a blank line; if we find a HEIGHT attribute, we may change
 *   this opinion.  
 */
enum vmconsole_html_br_mode
{
    HTML_DEFER_BR_NONE,                                  /* no pending <BR> */
    HTML_DEFER_BR_FLUSH,                       /* only need to flush output */
    HTML_DEFER_BR_BLANK                                /* need a blank line */
};



/* ------------------------------------------------------------------------ */
/*
 *   Output formatter stream interface.  A formatter stream performs
 *   formatting on a stream of displayed text.
 *   
 *   A given stream of display text might be fed into more than one
 *   formatter stream.  For example, if logging is turned on for a
 *   console, we'll feed the same text to the console's main formatting
 *   stream, which will end up being displayed to the user, and also to
 *   the log file's formatting stream, which will end up written out to
 *   the log file.
 *   
 *   Note that the formatter interface is internal to the console system.
 *   Client code should never need to refer to a formatter object, but
 *   should instead call the console object (CVmConsole).  
 */
typedef struct out_stream_info out_stream_info;
class CVmFormatter
{
public:
    CVmFormatter(class CVmConsole *console)
    {
        /* remember our display object */
        console_ = console;
    }

    virtual ~CVmFormatter() { }
    
    /* initialize */
    virtual void init()
    {
        /* start out at the first column */
        linepos_ = 0;
        linecol_ = 0;
        linebuf_[0] = '\0';

        /* start out at the first line */
        linecnt_ = 0;

        /* we're not in either "caps", "nocaps", or "allcaps" mode yet */
        capsflag_ = nocapsflag_ = allcapsflag_ = FALSE;

        /* 
         *   start in normal spacing mode - treat runs of whitespace as
         *   equivalent to a single space 
         */
        obey_whitespace_ = FALSE;

        /* we aren't previewing a line yet */
        preview_ = 0;

        /* we haven't flushed a new line yet */
        just_did_nl_ = FALSE;

        /* assume that the underlying system is not HTML-enabled */
        html_target_ = FALSE;

        /* presume this target accepts OS highlighting sequences */
        plain_text_target_ = FALSE;

        /* start out in text mode */
        html_mode_ = FALSE;

        /* start out in "normal" lexical state */
        html_mode_flag_ = HTML_MODE_NORMAL;

        /* not in an ignored tag yet (-> nesting depth is zero) */
        html_in_ignore_ = 0;

        /* not in title mode yet (-> nesting depth is zero) */
        html_in_title_ = 0;

        /* not yet deferring line breaks */
        html_defer_br_ = HTML_DEFER_BR_NONE;

        /* not yet in quotes (-> nesting depth is zero) */
        html_quote_level_ = 0;

        /* not in an ALT tag yet */
        html_allow_alt_ = FALSE;

        /* assume we're a normal display stream */
        is_disp_stream_ = TRUE;

        /* remember the OS line wrap setting */
        os_line_wrap_ = get_init_os_line_wrap();
    }

    /* get/set obey-whitespace mode */
    int get_obey_whitespace() const { return obey_whitespace_ != 0; }
    void set_obey_whitespace(int f) { obey_whitespace_ = (f != 0); }
    

    /* turn on CAPS mode */
    void caps()
    {
        /* turn on CAPS mode */
        capsflag_ = TRUE;

        /* turn off the NOCAPS and ALLCAPS modes */
        nocapsflag_ = FALSE;
        allcapsflag_ = FALSE;
    }

    /* turn on NOCAPS mode */
    void nocaps()
    {
        /* turn on NOCAPS mode */
        nocapsflag_ = TRUE;
        
        /* turn off CAPS and ALLCAPS mode */
        capsflag_ = FALSE;
        allcapsflag_ = FALSE;
    }
    
    /* turn on or off ALLCAPS mode */
    void allcaps(int all_caps)
    {
        /* set the ALLCAPS flag */
        allcapsflag_ = all_caps;
        
        /* clear the CAPS and NOCAPS flags */
        capsflag_ = FALSE;
        nocapsflag_ = FALSE;
    }

    /* 
     *   Display a string of a given byte-length.  The text is given in
     *   UTF-8 format.  We'll interpret embedded control codes (newlines,
     *   blank lines, caps flags, etc); depending on the display mode, we
     *   might also interpret HTML markup sequences.  
     */
    int format_text(VMG_ const char *s, size_t len);

    /* write a blank line to the stream */
    void write_blank_line(VMG0_);

    /*
     *   Flush the current line to the display, using the given type of
     *   line termination. 
     */
    void flush(VMG_ vm_nl_type nl);

    /*
     *   Get the escape character to put a stream into a given mode.
     *   Returns 0 if the underlying target doesn't support highlighting;
     *   we'll always return 0 if the underlying stream is a plain text
     *   target, otherwise we'll ask the OS layer for the highlighting
     *   escape code.  The mode codes are the same as for os_hilite().  
     */
    int get_hilite_char(int mode) const
    {
        return plain_text_target_ ? 0 : os_hilite(mode);
    }

    /* 
     *   Get the current HTML mode.  This returns true if we've processed
     *   a \H+ sequence and have not encountered a \H- sequence, so that
     *   we're processing HTML markups (either directly in the formatter,
     *   or in the underlying renderer).  
     */
    int is_html_mode() const { return html_mode_; }

    /* 
     *   determine if I'm an HTML target - this returns true if the
     *   underyling renderer accepts HTML, in which case we write HTML
     *   markups directly to the os_xxx routines (os_printf, for example).
     */
    int is_html_target() const { return html_target_; }

    /* reset the MORE prompt line counter */
    void reset_line_count() { linecnt_ = 0; }

    /* 
     *   Low-level display routine.  The text passed to this routine is in
     *   the local character set, so no conversions are necessary for
     *   calling the os_xxx routines.  
     */
    virtual void disp_text_local_cs(const char *txt) = 0;

protected:
    /*
     *   Write a line (or a partial line) of text to the stream, using the
     *   indicated line breaking.  The text is given as wide Unicode
     *   characters.  
     */
    void write_text(VMG_ const wchar_t *txt, vm_nl_type nl);

    /* write a tab to the stream */
    void write_tab(VMG0_);

    /* flush the current line, with or without padding */
    void flush_line(VMG_ int padding);

    /* 
     *   Buffer a character of output.  The character is presented to us
     *   as a wide Unicode character.  
     */
    void buffer_char(VMG_ wchar_t c);

    /* 
     *   Buffer a string of output to the stream.  The string is in UTF-8
     *   format. 
     */
    void buffer_string(VMG_ const char *txt);

    /* buffer a wide unicode character string */
    void buffer_wstring(VMG_ const wchar_t *txt);

    /* get the next wide unicode character in a UTF8-encoded string */
    static wchar_t next_wchar(const char **s, size_t *len);

    /* service routine - translate a wide character string to an integer */
    static int wtoi(const wchar_t *p);

    /* 
     *   Determine if we should use MORE mode in the formatter layer; if
     *   this returns true, we handle MORE prompting ourselves, otherwise
     *   we handle it through the OS layer.
     *   
     *   The default version of this function is implemented in the output
     *   formatter configuration module.  This function can also be
     *   overridden in a subclass (for example, a log stream never uses
     *   MORE mode, no matter what configuration we're using).  
     */
    virtual int use_more_mode() const;

    /* 
     *   Determine if this output stream uses OS-level line wrapping - if
     *   this is returns true, the output formatter will not insert a
     *   newline at the end of a line that it's flushing for word
     *   wrapping, but will instead let the underlying OS display layer
     *   handle the wrapping.  (We cache this value locally because we
     *   check it frequently.)
     *   
     *   This is implemented in the formatter configuration module.  
     */
    int get_init_os_line_wrap() const;

    /*
     *   Determine if the underlying stream interprets HTML markups.  We
     *   call this during initialization to set our html_target_ member
     *   (we cache the result since we check it frequently).
     *   
     *   This is implemented in the formatter configuration module.  
     */
    int get_init_html_target() const;

    /*
     *   Turn HTML mode on or off in the underlying OS-level HTML renderer.
     *   
     *   This is implemented in the formatter configuration module. 
     */
    void start_html();
    void end_html();

    /* the console that owns this formatter stream */
    class CVmConsole *console_;
    
    /* current line position and output column */
    int linepos_;
    int linecol_;

    /* 
     *   "preview" state for line flushing - this is the last column at
     *   which we flushed a partial line 
     */
    int preview_;

    /* number of lines on the screen (since last MORE prompt) */
    int linecnt_;

    /* 
     *   Output buffer.  We keep the output buffer as wide Unicode
     *   characters, and translate to the local character set when we
     *   flush the buffer and send the text to the os_xxx routines for
     *   display. 
     */
    wchar_t linebuf_[OS_MAXWIDTH];

    /* obey-whitespace mode - treat whitespace as literal */
    unsigned int obey_whitespace_ : 1;

    /* CAPS mode - next character output is converted to upper-case */
    unsigned int capsflag_ : 1;

    /* NOCAPS mode - next character output is converted to lower-case */
    unsigned int nocapsflag_ : 1;

    /* ALLCAPS mode - all characters output are converted to upper-case */
    unsigned int allcapsflag_ : 1;

    /* flag indicating that we just flushed a new line */
    unsigned int just_did_nl_ : 1;

    /*
     *   Flag indicating that the underlying output system wants to
     *   receive its output as HTML.
     *   
     *   If this is true, we'll pass through HTML to the underlying output
     *   system, and in addition generate HTML sequences for certain
     *   TADS-native escapes (for example, we'll convert the "\n" sequence
     *   to a <BR> sequence).
     *   
     *   If this is false, we'll do just the opposite: we'll remove HTML
     *   from the output stream and convert it into normal text sequences. 
     */
    unsigned int html_target_ : 1;

    /*
     *   Flag indicating that the target uses plain text.  If this flag is
     *   set, we won't add the OS escape codes for highlighted characters. 
     */
    unsigned int plain_text_target_ : 1;

    /* 
     *   flag: the underlying OS layer handles line wrapping, so we never
     *   need to write newlines when flushing our line buffer except when
     *   we want to indicate a hard line break 
     */
    unsigned int os_line_wrap_ : 1;
    
    /* 
     *   Flag indicating that the caller is displaying HTML.  We always
     *   start off in text mode; the client can switch to HTML mode by
     *   displaying a special escape sequence, and can switch back to text
     *   mode by displaying a separate special escape sequence.  
     */
    unsigned int html_mode_ : 1;

    /* current lexical analysis mode */
    vmconsole_html_mode html_mode_flag_;

    /* <BR> defer mode */
    vmconsole_html_br_mode html_defer_br_;

    /* 
     *   HTML "ignore" mode - we suppress all output when parsing the
     *   contents of a <TITLE> or <ABOUTBOX> tag.  This is a counter that
     *   keeps track of the nesting depth for ignored tags.  
     */
    int html_in_ignore_;

    /*
     *   HTML <TITLE> mode - when we're in this mode, we're gathering the
     *   title (i.e., we're inside a <TITLE> tag's contents).  We'll copy
     *   characters to the title buffer rather than the normal output
     *   buffer, and then call os_set_title() when we reach the </TITLE>
     *   tag.  This is a counter that keeps track of the nesting depth of
     *   <TITLE> tags.  
     */
    int html_in_title_;

    /* buffer for the title */
    char html_title_buf_[HTML_TITLE_BUF_SIZE];

    /* pointer to next available character in title buffer */
    char *html_title_ptr_;

    /* 
     *   quoting level - this is a counter that keeps track of the nesting
     *   depth of <Q> tags 
     */
    int html_quote_level_;

    /*
     *   Parsing mode flag for ALT attributes.  If we're parsing a tag
     *   that allows ALT, such as IMG or SOUND, we'll set this flag, then
     *   insert the ALT text if we encounter it during parsing.  
     */
    unsigned int html_allow_alt_ : 1;

    /* 
     *   Flag: this is a display stream.  Other types of streams (such as
     *   log file streams) should set this to false. 
     */
    unsigned int is_disp_stream_ : 1;
};

/* ------------------------------------------------------------------------ */
/*
 *   Formatter subclass for the main display 
 */
class CVmFormatterDisp: public CVmFormatter
{
public:
    CVmFormatterDisp(class CVmConsole *console)
        : CVmFormatter(console)
    {
    }
    
    /* initialize */
    void init()
    {
        /* inherit base class initialization */
        CVmFormatter::init();
        
        /* 
         *   if we're compiled for HTML mode, set the standard output
         *   stream so that it knows it has an HTML target - this will
         *   ensure that HTML tags are passed through to the underlying
         *   stream, and that we generate HTML equivalents for our own
         *   control sequences 
         */ 
        html_target_ = get_init_html_target();
    }
    
    /* display text */
    virtual void disp_text_local_cs(const char *txt)
    {
        /* display the text through the OS interface */
        os_printf("%s", txt);
    }
    
protected:
};

/* ------------------------------------------------------------------------ */
/*
 *   Formatter subclass for the log file 
 */
class CVmFormatterLog: public CVmFormatter
{
    friend class CVmConsole;
    
public:
    CVmFormatterLog(class CVmConsole *console)
        : CVmFormatter(console)
    {
    }
    
    ~CVmFormatterLog()
    {
        /* close any active log file */
        close_log_file();
    }
    
    /* initialize */
    void init()
    {
        /* inherit base class initialization */
        CVmFormatter::init();
        
        /* use plain text in the log file stream */
        plain_text_target_ = TRUE;
        
        /* we're not a display stream */
        is_disp_stream_ = FALSE;
        
        /* no log file yet */
        logfp_ = 0;
    }
    
    /* display text to the underlying OS device */
    virtual void disp_text_local_cs(const char *txt)
    {
        /* display the text through the log file */
        if (logfp_ != 0)
            os_fprintf(logfp_, "%s", txt);
    }

    /* don't use MORE mode in a log stream */
    int use_more_mode() const { return FALSE; }
    
protected:
    /* open a log file */
    int open_log_file(const char *fname, int initial_html_mode);
    
    /* close the log file */
    int close_log_file();
    
    /* my log file handle */
    osfildef *logfp_;
    
    /* log file name */
    char logfname_[OSFNMAX];
};


#endif /* VMCONSOL_H */
