#ifdef RCSID
static char RCSid[] =
"$Header$";
#endif

/* Copyright (c) 1999 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  vmmain.cpp - T3 VM main entrypoint - execute an image file
Function
  Main entrypoint for executing a T3 image file.  Loads an image file
  and begins execution.  Returns when execution terminates.
Notes
  
Modified
  10/07/99 MJRoberts  - Creation
*/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "os.h"
#include "t3std.h"
#include "vmerr.h"
#include "vmfile.h"
#include "vmimage.h"
#include "vmrun.h"
#include "vmimgrb.h"
#include "vmmain.h"
#include "vmhost.h"
#include "vmhostsi.h"
#include "vminit.h"
#include "vmconsol.h"
#include "vmpredef.h"
#include "vmobj.h"
#include "vmvsn.h"
#include "charmap.h"
#include "vmsave.h"
#include "vmtype.h"


/* ------------------------------------------------------------------------ */
/*
 *   Execute an image file.  If an exception occurs, we'll display a
 *   message on the console, and we'll return the error code; we'll return
 *   zero on success.  If an error occurs, we'll fill in 'errbuf' with a
 *   message describing the problem.  
 */
int vm_run_image(CVmMainClientIfc *clientifc,
                 const char *image_file_name,
                 class CVmHostIfc *hostifc,
                 const char *const *prog_argv, int prog_argc,
                 const char *script_file, const char *log_file,
                 const char *cmd_log_file,
                 int load_from_exe, int show_banner, const char *charset,
                 const char *saved_state)
{
    CVmFile *fp = 0;
    CVmImageLoader *volatile loader = 0;
    CVmImageFile *volatile imagefp = 0;
    int retval;
    vm_globals *vmg__;

    /* presume we will return success */
    retval = 0;

    /* create the file object */
    fp = new CVmFile();

    /* initialize the VM */
    vm_initialize(&vmg__, hostifc, charset);

    /* set up scripting if desired */
    if (script_file != 0)
        G_console->open_script_file(script_file, TRUE, FALSE);

    /* set up logging if desired */
    if (log_file != 0)
        G_console->open_log_file(log_file);

    /* set up command logging if desired */
    if (cmd_log_file != 0)
        G_console->open_command_log(cmd_log_file);

    /* show the banner if desired */
    if (show_banner)
    {
        G_console->format_text(vmg_ T3VM_BANNER_STRING);
        G_console->write_blank_line(vmg0_);
    }

    err_try
    {
        if (load_from_exe)
        {
            osfildef *exe_fp;
            
            /* find the image within the executable */
            exe_fp = os_exeseek(image_file_name, "TGAM");
            if (exe_fp == 0)
                err_throw(VMERR_NO_IMAGE_IN_EXE);

            /* 
             *   set up to read from the executable at the location of the
             *   embedded image file that we just found 
             */
            fp->set_file(exe_fp, osfpos(exe_fp));
        }
        else
        {
            /* reading from a normal file - open the file */
            fp->open_read(image_file_name, OSFTT3IMG);
        }

        /* create the loader */
        imagefp = new CVmImageFileExt(fp);
        loader = new CVmImageLoader(imagefp, image_file_name);

        /* load the image */
        loader->load(vmg0_);

        /* let the client prepare for execution */
        clientifc->pre_exec(VMGLOB_ADDR);

        /*
         *   If we're restoring a saved state, restore the saved state
         *   file and use it to resume execution.  Otherwise, begin
         *   execution at the main entrypoint.  
         */
        if (saved_state != 0)
        {
            /* run the program from the saved state */
            loader->run_restore(vmg_ saved_state);
        }
        else
        {
            /* run the program from the main entrypoint */
            loader->run(vmg_ prog_argv, prog_argc);
        }

        /* tell the client we're done with execution */
        clientifc->post_exec(VMGLOB_ADDR);
    }
    err_catch(exc)
    {
        char errbuf[512];

        /* tell the client execution failed due to an error */
        clientifc->post_exec_err(VMGLOB_ADDR);

        /* note the error code for returning to the caller */
        retval = exc->get_error_code();
        
        /* check for unhandled program exceptions */
        if (exc->get_error_code() == VMERR_UNHANDLED_EXC)
        {
            const char *msg;
            size_t msg_len;
            
            /* 
             *   This is not a VM error, but is simply an exception that
             *   the program itself threw but did not handle.  We might be
             *   able to find an informational message in the exception
             *   object itself.  
             */

            /* get the exception's message, if available */
            msg = CVmRun::get_exc_message(vmg_ exc, &msg_len);
            if (msg != 0)
            {
                /* use the message that comes with the exception */
                sprintf(errbuf, "Unhandled exception: %.*s",
                        (int)msg_len, msg);
            }
            else
            {
                /* no message - display a generic exception message */
                strcpy(errbuf, "Unhandled exception");
            }
        }
        else
        {
            const char *msg;

            /* 
             *   It's a VM exception, so we can determine the error's
             *   meaning from the error code.  Look up the message for the
             *   error code in our error message list.  
             */
            msg = err_get_msg(vm_messages, vm_message_count,
                              exc->get_error_code(), FALSE);

            /* if that failed, just show the error number */
            if (msg == 0)
            {
                /* no message - just show the error code */
                sprintf(errbuf, "VM Error: code %d", exc->get_error_code());
            }
            else
            {
                char buf[512];

                /* format the message from the exception parameters */
                err_format_msg(buf, sizeof(buf), msg, exc);

                /* add a prefix */
                sprintf(errbuf, "VM Error: %s", buf);
            }
        }

        /* flush the console */
        G_console->flush(vmg0_);

        /* display the message */
        clientifc->display_error(VMGLOB_ADDR, errbuf);
    }
    err_end;

    /* flush any pending buffered output */
    G_console->flush(vmg_ VM_NL_NONE);

    /* unload the image */
    if (loader != 0)
        loader->unload(vmg0_);

    /* delete the loader and the image file object */
    if (loader != 0)
        delete loader;
    if (imagefp != 0)
        delete imagefp;

    /* terminate the VM */
    vm_terminate(vmg__);

    /* delete the file */
    if (fp != 0)
        delete fp;

    /* return the status code */
    return retval;
}

/* ------------------------------------------------------------------------ */
/*
 *   Main Entrypoint for command-line invocations.  For simplicity, a
 *   normal C main() or equivalent entrypoint can invoke this routine
 *   directly, using the usual argc/argv conventions.
 *   
 *   Returns a status code suitable for use with exit(): OSEXSUCC if we
 *   successfully loaded and ran an executable, OSEXFAIL on failure.  If
 *   an error occurs, we'll fill in 'errbuf' with a message describing the
 *   problem.  
 */
int vm_run_image_main(CVmMainClientIfc *clientifc,
                      const char *executable_name,
                      int argc, char **argv, int defext,
                      CVmHostIfc *hostifc)
{
    int curarg;
    char image_file_name[OSFNMAX];
    int created_hostifc;
    int stat;
    const char *script_file;
    const char *log_file;
    const char *cmd_log_file;
    int load_from_exe;
    int show_banner;
    int found_image;
    int hide_usage;
    int usage_err;
    const char *charset;
    char *saved_state;

    /* we haven't found an image file yet */
    found_image = FALSE;

    /* presume we'll show usage on error */
    hide_usage = FALSE;

    /* presume there will be no usage error */
    usage_err = FALSE;

    /* presume we won't have any console input/output files */
    script_file = 0;
    log_file = 0;
    cmd_log_file = 0;

    /* presume we'll use the default OS character set */
    charset = 0;

    /* presume we won't show the banner */
    show_banner = FALSE;

    /* presume we won't load from the .exe file */
    load_from_exe = FALSE;

    /* presume we won't restore a saved state file */
    saved_state = 0;

    /* create our host interface if none was provided */
    if (hostifc == 0)
    {
        /* create our own interface */
        hostifc = new CVmHostIfcStdio(argv[0]);

        /* note that we created it */
        created_hostifc = TRUE;
    }
    else
    {
        /* we didn't create the host interface */
        created_hostifc = FALSE;
    }

    /* scan options */
    for (curarg = 1 ; curarg < argc && argv[curarg][0] == '-' ; ++curarg)
    {
        /* check the argument */
        switch(argv[curarg][1])
        {
        case 'b':
            if (strcmp(argv[curarg], "-banner") == 0)
            {
                /* make a note to show the banner */
                show_banner = TRUE;
            }
            else
                goto opt_error;
            break;

        case 'c':
            if (argv[curarg][2] == 's')
            {
                ++curarg;
                if (curarg < argc)
                    charset = argv[curarg];
                else
                    goto opt_error;
            }
            else
                goto opt_error;
            break;

        case 'n':
            if (strcmp(argv[curarg], "-nobanner") == 0)
            {
                /* make a note not to show the banner */
                show_banner = FALSE;
            }
            else
                goto opt_error;
            break;
            
        case 's':
            /* file safety level - check the range */
            if (argv[curarg][2] < '0' || argv[curarg][2] > '4')
            {
                /* invalid level */
                goto opt_error;
            }
            else
            {
                /* set the level in the host application */
                hostifc->set_io_safety(argv[curarg][2] - '0');
            }
            break;

        case 'i':
            /* read from a script file - the next argument is the filename */
            ++curarg;
            if (curarg < argc)
                script_file = argv[curarg];
            else
                goto opt_error;
            break;

        case 'l':
            /* log output to file */
            ++curarg;
            if (curarg < argc)
                log_file = argv[curarg];
            else
                goto opt_error;
            break;

        case 'o':
            /* log commands to file */
            ++curarg;
            if (curarg < argc)
                cmd_log_file = argv[curarg];
            else
                goto opt_error;
            break;

        case 'p':
            /* check what follows */
            if (strcmp(argv[curarg], "-plain") == 0)
            {
                /* set plain ASCII mode at the OS level */
                os_plain();
                break;
            }
            else
                goto opt_error;
            break;

        case 'r':
            /* get the name of the saved state file to restore */
            ++curarg;
            if (curarg < argc)
                saved_state = argv[curarg];
            else
                goto opt_error;
            break;
            
        default:
        opt_error:
            /* discard remaining arguments */
            curarg = argc;

            /* note the error */
            usage_err = TRUE;
            break;
        }
    }

    /* 
     *   If there was no usage error so far, but we don't have an image
     *   filename argument, try to find the image file some other way.
     */
    if (usage_err)
    {
        /* there was a usage error - don't bother looking for an image file */
    }
    else if (curarg + 1 <= argc)
    {
        /* the last argument is the image file name */
        strcpy(image_file_name, argv[curarg]);
        found_image = TRUE;

        /* 
         *   If the given filename exists, use it as-is; otherwise, if
         *   we're allowed to add an extension, try applying a default
         *   extension of "t3x" to the given name.  
         */
        if (defext && osfacc(image_file_name))
        {
            /* the given name doesn't exist - try a default extension */
            os_defext(image_file_name, "t3x");
        }
    }
    else
    {
        osfildef *fp;
            
        /* look for an image file attached to the executable */
        fp = os_exeseek(argv[0], "TGAM");
        if (fp != 0)
        {
            /* close the file */
            osfcls(fp);

            /* note that we want to load from the executable */
            load_from_exe = TRUE;

            /* use the executable filename as the image file */
            strcpy(image_file_name, argv[0]);
            found_image = TRUE;
        }

        /* 
         *   if we still haven't found an image file, try to get the image
         *   file from the saved state file, if one was specified 
         */
        if (!found_image && saved_state != 0)
        {
            osfildef *save_fp;

            /* open the saved state file */
            save_fp = osfoprb(saved_state, OSFTT3SAV);
            if (save_fp != 0)
            {
                /* get the name of the image file */
                if (CVmSaveFile::restore_get_image(
                    save_fp, image_file_name, sizeof(image_file_name)) == 0)
                {
                    /* we successfully obtained the filename */
                    found_image = TRUE;
                }

                /* close the file */
                osfcls(save_fp);
            }
        }

        /* 
         *   if we haven't found the image, and the host system provides a
         *   way of asking the user for a filename, try that 
         */
        if (!found_image)
        {
            /* ask the host system for a game name */
            switch (hostifc->get_image_name(image_file_name,
                                            sizeof(image_file_name)))
            {
            case VMHOST_GIN_IGNORED:
                /* no effect - we have no new information */
                break;

            case VMHOST_GIN_CANCEL:
                /* 
                 *   the user cancelled the dialog - we don't have a
                 *   filename, but we also don't want to show usage, since
                 *   the user chose not to proceed 
                 */
                hide_usage = TRUE;
                break;
                
            case VMHOST_GIN_ERROR:
                /* 
                 *   an error occurred showing the dialog - there's not
                 *   much we can do except show the usage message 
                 */
                break;

            case VMHOST_GIN_SUCCESS:
                /* that was successful - we have an image file now */
                found_image = TRUE;
                break;
            }
        }
    }

    /* 
     *   if we don't have an image file name by this point, we can't
     *   proceed - show the usage message and terminate 
     */
    if (usage_err || !found_image)
    {
        char buf[OSFNMAX + 1024];
        vm_globals *vmg__;

        /* show the usage message if allowed */
        if (!hide_usage)
        {
            /* build the usage message */
            sprintf(buf,
                    "%s\n"
                    "usage: %s [options] <image-file-name> [arguments]\n"
                    "options:\n"
                    "  -banner - show the version/copyright banner\n"
                    "  -cs xxx - use character set 'xxx' for keyboard "
                    "and display\n"
                    "  -i file - read command line input from file\n"
                    "  -l file - log all console input/output to file\n"
                    "  -o file - log console input to file\n"
                    "  -plain  - run in plain mode (no cursor positioning, "
                    "colors, etc.)\n"
                    "  -r file - restore saved state from file\n"
                    "  -s#     - set I/O safety level (# in range 0 to 4 - 0 "
                    "is the least\n"
                    "            restrictive, 4 allows no file I/O at all)\n"
                    "\n"
                    "If provided, the optional arguments after the image "
                    "file name are passed\n"
                    "to the program's main entrypoint.\n",
                    T3VM_BANNER_STRING, executable_name);
            
            /* initialize the VM (so we have a console) */
            vm_initialize(&vmg__, hostifc, 0);
            
            /* display the message */
            clientifc->display_error(VMGLOB_ADDR, buf);
            
            /* terminate the VM */
            vm_terminate(vmg__);
        }
        
        /* delete the host interface if necessary */
        if (created_hostifc)
            delete hostifc;
        
        /* return failure */
        return OSEXFAIL;
    }

    /* run the program */
    stat = vm_run_image(clientifc, image_file_name, hostifc,
                        argv + curarg, argc - curarg,
                        script_file, log_file, cmd_log_file,
                        load_from_exe, show_banner, charset,
                        saved_state);

    /* delete our host interface if we created it */
    if (created_hostifc)
        delete hostifc;

    /* return the status code */
    return stat;
}
