#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/tads3/VMTOBJ.CPP,v 1.3 1999/05/17 02:52:28 MJRoberts Exp $";
#endif

/* Copyright (c) 1998 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  vmtobj.cpp - TADS object implementation
Function
  
Notes
  
Modified
  10/30/98 MJRoberts  - Creation
*/

#include <stdlib.h>
#include <assert.h>

#include "t3std.h"
#include "vmglob.h"
#include "vmerr.h"
#include "vmerrnum.h"
#include "vmobj.h"
#include "vmtobj.h"
#include "vmundo.h"
#include "vmtype.h"
#include "vmfile.h"
#include "vmstack.h"
#include "vmrun.h"
#include "vmpredef.h"


/* ------------------------------------------------------------------------ */
/*
 *   statics 
 */

/* metaclass registration object */
static CVmMetaclassTads metaclass_reg_obj;
CVmMetaclass *CVmObjTads::metaclass_reg_ = &metaclass_reg_obj;


/* ------------------------------------------------------------------------ */
/*
 *   Static creation methods 
 */

/* create dynamically using stack arguments */
vm_obj_id_t CVmObjTads::create_from_stack(VMG_ const uchar **pc_ptr,
                                          uint argc)
{
    vm_obj_id_t id;
    CVmObjTads *obj;
    vm_val_t val;
    vm_obj_id_t srcobj;

    /* make sure we have at least one argument */
    if (argc < 1)
        err_throw(VMERR_WRONG_NUM_OF_ARGS);

    /* pop the superclass argument - it must be an object or nil */
    G_stk->pop(&val);
    if (val.typ != VM_OBJ && val.typ != VM_NIL)
        err_throw(VMERR_OBJ_VAL_REQD_SC);

    /* if it's the invalid object, treat it as nil */
    if (val.typ == VM_OBJ && val.val.obj == VM_INVALID_OBJ)
        val.set_nil();

    /* count the removal of the first argument */
    --argc;

    /* 
     *   create the object - this type of construction is never used for
     *   root set objects 
     */
    id = vm_new_id(vmg_ FALSE);

    /* 
     *   create a TADS object with the appropriate number of superclasses
     *   (0 if no superclass was specified, 1 if one was), and the default
     *   number of initial mutable properties 
     */
    obj = new (vmg_ id) CVmObjTads(vmg_ (val.typ == VM_NIL ? 0 : 1),
                                   VMTOBJ_PROP_INIT);

    /* set the object's superclass */
    if (val.typ != VM_NIL)
        obj->set_sc(0, val.val.obj);

    /* 
     *   Invoke the object's "construct" method, passing it the arguments
     *   that are still on the stack.  If the new object doesn't define or
     *   inherit the "construct" method, simply push the new object
     *   reference onto the stack directly.  
     */
    if (obj->get_prop(vmg_ G_predef->obj_construct, &val, id, &srcobj, 0))
    {
        vm_val_t srcobj_val;
        vm_val_t id_val;
        
        /* 
         *   A "construct" method is defined - have the interpreter invoke
         *   it, which will set up the interpreter to start executing its
         *   byte-code.  This is all we need to do, since we assume and
         *   require that the constructor will return the new object as
         *   its return value when it's done.  
         */
        srcobj_val.set_obj(srcobj);
        id_val.set_obj(id);
        G_interpreter->get_prop(vmg_ pc_ptr, &srcobj_val,
                                G_predef->obj_construct, &id_val, argc);
    }
    else
    {
        /* 
         *   there's no "construct" method defined - if we have any
         *   arguments, its an error 
         */
        if (argc != 0)
            err_throw(VMERR_WRONG_NUM_OF_ARGS);

        /* leave the new object value in R0 */
        G_interpreter->get_r0()->set_obj(id);
    }

    /* return the new object */
    return id;
}

/* create an object with no initial extension */
vm_obj_id_t CVmObjTads::create(VMG_ int in_root_set)
{
    vm_obj_id_t id = vm_new_id(vmg_ in_root_set);
    new (vmg_ id) CVmObjTads();
    return id;
}

/* 
 *   Create an object with a given number of superclasses, and a given
 *   property table size.  Each superclass must be set before the object
 *   can be used, and the property table is initially empty.
 *   
 *   This form is used to create objects dynamically; this call is never
 *   used to load an object from an image file.  
 */
vm_obj_id_t CVmObjTads::create(VMG_ int in_root_set,
                               ushort superclass_count, ushort prop_count)
{
    vm_obj_id_t id = vm_new_id(vmg_ in_root_set);
    new (vmg_ id) CVmObjTads(vmg_ superclass_count, prop_count);
    return id;
}


/* ------------------------------------------------------------------------ */
/*
 *   Constructors 
 */

/*
 *   Create an object with a given number of superclasses, and a given
 *   property table size.  The superclasses must be individually set
 *   before the object can be used, and the property table is initially
 *   empty.
 *   
 *   This constructor is used only when creating a new object dynamically,
 *   and is never used to load an object from an image file.  
 */
CVmObjTads::CVmObjTads(VMG_ ushort superclass_count, ushort prop_count)
{
    /* allocate the memory */
    alloc_mem(vmg_ superclass_count, prop_count, FALSE);
}


/* ------------------------------------------------------------------------ */
/*
 *   receive notification of deletion 
 */
void CVmObjTads::notify_delete(VMG_ int /*in_root_set*/)
{
    /* free our extension */
    if (ext_ != 0)
        G_mem->get_var_heap()->free_mem(ext_);
}

/* ------------------------------------------------------------------------ */
/*
 *   Create an instance of this class 
 */
void CVmObjTads::create_instance(VMG_ vm_obj_id_t self,
                                 const uchar **pc_ptr, uint argc) 
{
    vm_val_t val;

    /* push myself as the superclass */
    val.set_obj(self);
    G_stk->push(&val);

    /* use the normal stack creation routine */
    create_from_stack(vmg_ pc_ptr, argc+1);
}

/* ------------------------------------------------------------------------ */
/*
 *   Determine if the object has a finalizer method 
 */
int CVmObjTads::has_finalizer(VMG_ vm_obj_id_t self)
{
    vm_val_t val;
    vm_obj_id_t srcobj;

    /* 
     *   look up the finalization method - if it's defined, and it's a
     *   method, invoke it; otherwise do nothing 
     */
    return (G_predef->obj_destruct != VM_INVALID_PROP
            && get_prop(vmg_ G_predef->obj_destruct, &val, self, &srcobj, 0)
            && (val.typ == VM_CODEOFS || val.typ == VM_NATIVE_CODE));
}

/* ------------------------------------------------------------------------ */
/*
 *   Invoke the object's finalizer 
 */
void CVmObjTads::invoke_finalizer(VMG_ vm_obj_id_t self)
{
    vm_val_t val;
    vm_obj_id_t srcobj;

    /* 
     *   look up the finalization method - if it's defined, and it's a
     *   method, invoke it; otherwise do nothing 
     */
    if (G_predef->obj_destruct != VM_INVALID_PROP
        && get_prop(vmg_ G_predef->obj_destruct, &val, self, &srcobj, 0)
        && (val.typ == VM_CODEOFS || val.typ == VM_NATIVE_CODE))
    {
        /* 
         *   invoke the finalizer in a protected frame, to ensure that we
         *   catch any exceptions that are thrown out of the finalizer 
         */
        err_try
        {
            vm_val_t srcobj_val;
            vm_val_t self_val;
            
            /* 
             *   Invoke the finalizer.  Use a recursive VM invocation,
             *   since the VM must return to the garbage collector, not to
             *   what it was doing in the enclosing stack frame. 
             */
            srcobj_val.set_obj(srcobj);
            self_val.set_obj(self);
            G_interpreter->get_prop(vmg_ 0, &srcobj_val,
                                    G_predef->obj_destruct, &self_val, 0);
        }
        err_catch(exc)
        {
            /* silently ignore the error */
        }
        err_end;
    }
}


/* ------------------------------------------------------------------------ */
/*
 *   Calculate how much space we need for the object extension 
 */
size_t CVmObjTads::calc_alloc_size(ushort superclass_count,
                                   ushort mod_prop_count,
                                   int from_image) const
{
    size_t siz;
        
    /*
     *   Calculate how much space we need.  We need space for our header;
     *   the superclass count; the property table slot count; the
     *   superclass slots, each of which requires an object ID; the
     *   property slots, each of which requires a property ID and a data
     *   holder; and the undo bit array, which consists of one bit per
     *   property, packed and rounded up to a whole byte 
     */
    siz = sizeof(vm_tadsobj_hdr)                                  /* header */
          + (mod_prop_count * VMTOBJ_PROP_SLOT_SIZE) /* modified prop slots */
          + ((mod_prop_count + 7) >> 3);                  /* undo bit array */

    /*
     *   If this isn't being loaded from the image file, add in the space
     *   we'll need for the data that would normally go in the header,
     *   which we need to store in the extension since we don't have the
     *   image data to hold it.  The header data consists of the
     *   superclass count (a UINT2), the load-image property count (a
     *   UINT2), the flags (a UINT2), plus the superclasses (an OBJECT_ID
     *   per).  
     */
    if (!from_image)
        siz += 3*VMB_UINT2 + (superclass_count * VMB_OBJECT_ID);

    /* return the size */
    return siz;
}


/* ------------------------------------------------------------------------ */
/*
 *   Set a property 
 */
void CVmObjTads::set_prop(VMG_ CVmUndo *undo, vm_obj_id_t self,
                          vm_prop_id_t prop, const vm_val_t *val)
{
    char *slotptr;
    vm_val_t oldval;
    int found;

    /* 
     *   Look for a modified property slot containing this property.  If
     *   the property isn't among the modified slots, we will simply add
     *   it as a new modified slot; otherwise, we must change the existing
     *   slot.
     *   
     *   Note that it doesn't matter whether or not the property exists
     *   among the permanent slots - since we're making a change, we're
     *   only interested in the modifiable slots.  
     */
    slotptr = look_up_mod_prop(prop, TRUE);

    /* determine whether we found the property */
    found = (slotptr != 0 && get_slotptr_propid(slotptr) != VM_INVALID_PROP);

    /* 
     *   Get the old value - if we didn't find the property, record the
     *   old value as 'empty', so that we know to delete it if we should
     *   have to undo this change.
     */
    if (found)
        get_slotptr_val(slotptr, &oldval);
    else
        oldval.set_empty();

    /*
     *   If we already have undo for this property for the current
     *   savepoint, as indicated by the undo bit for the property, we
     *   don't need to save undo for this change, since we already have an
     *   undo record in the current savepoint.  Otherwise, we need to add
     *   an undo record for this savepoint.  Note that we need to create
     *   undo if we don't have a slot pointer, since this means we'll be
     *   adding a new property -- since the property wasn't here before,
     *   it didn't have a slot, hence there's no way we could have undo
     *   for it.  
     */
    if (undo != 0
        && (!found || !get_undo_bit(get_undo_slotnum(slotptr))))
        undo->add_new_record_prop_key(vmg_ self, prop, &oldval);

    /*
     *   If we didn't find a slot to contain the property, expand the
     *   table 
     */
    if (slotptr == 0)
    {
        size_t old_mod_slots;

        /* remember the current count */
        old_mod_slots = get_mod_prop_alloc();
        
        /* allocate additional property slots */
        expand_prop_table(vmg_ get_mod_prop_alloc() + VMTOBJ_PROP_INC,
                          vm_objp(vmg_ self));

        /* get a pointer to the first new property slot */
        slotptr = get_mod_prop_slot_ptr(old_mod_slots);
    }
        
    /* store the property */
    set_prop_slotptr(slotptr, prop, val);

    /* 
     *   if we just filled in any empty slot, and there's anything
     *   allocated in the table beyond this slot, make sure the next slot
     *   is marked as empty - we need to do this now, since we don't
     *   preinitialize all of the slots when we allocate them 
     */
    if (!found
        && get_mod_slot_from_slot_ptr(slotptr) + 1 < get_mod_prop_alloc())
    {
        char *slotptr_nxt;
        
        /* move on to the next slot */
        slotptr_nxt = slotptr;
        inc_prop_slot_ptr(&slotptr_nxt);

        /* mark it empty */
        set_slotptr_propid(slotptr_nxt, VM_INVALID_PROP);
    }

    /* set the undo bit for the slot */
    if (undo != 0)
        set_undo_bit(get_undo_slotnum(slotptr));
}


/* ------------------------------------------------------------------------ */
/*
 *   Expand the property table to make room for new properties.  
 */
void CVmObjTads::expand_prop_table(VMG_ ushort new_mod_slots,
                                   CVmObject *obj)
{
    size_t siz;
    uchar *old_undo_bits;
    ushort old_mod_slots;
    int in_image;
    size_t copy_size;

    /* 
     *   note whether the object is in the image file or not (we must do
     *   this before we reallocate the block, because things will get
     *   inconsistent once we do so - we need to know now, while things
     *   are still consistent, so that we can make things consistent once
     *   more after reallocating) 
     */
    in_image = is_from_image();

    /* get the current number of modifiable slots */
    old_mod_slots = get_mod_prop_alloc();

    /* 
     *   if we already have the requested number of slots available,
     *   there's nothing we need to do 
     */
    if (old_mod_slots >= new_mod_slots)
        return;

    /* 
     *   calculate how much memory we need to expand the modifiable
     *   property list to the requested size 
     */
    siz = calc_alloc_size(get_sc_count(), new_mod_slots, in_image);
    
    /* reallocate our memory at the new, larger size */
    ext_ = (char *)G_mem->get_var_heap()->realloc_mem(siz, ext_, obj);

    /* 
     *   If we're not loaded from an image file, set our fake-image-data
     *   pointer to point to the extension following the variable part -
     *   calculate this by adding to the extension base pointer the size
     *   we'd allocate if we actually did have the data in the image file 
     */
    if (!in_image)
    {
        /* 
         *   set our fake image-data pointer by adding the size that we'd
         *   have if we actually got that data from the image file to our
         *   own memory pointer 
         */
        get_hdr()->image_data = 
            ext_ + calc_alloc_size(0, get_mod_prop_alloc(), TRUE);
    }

    /* remember the old undo bit pointer in the new allocation block */
    old_undo_bits = get_undo_bits();

    /* set the new, expanded size */
    set_mod_prop_alloc(new_mod_slots);

    /* 
     *   Copy the old undo bits to their new location -- the old undo bit
     *   list will always be smaller than the new undo bit list, since
     *   we've added more modifiable properties; hence, we simply copy the
     *   old undo bit list size.
     */
    copy_size = ((old_mod_slots + 7) >> 3);

    /* copy the data */
    memmove(get_undo_bits(), old_undo_bits, copy_size);

    /* 
     *   Set the first new slot to empty.  Note - it's important that we
     *   wait until AFTER we've copied the undo bits to initialize the new
     *   property slot, because the new property slot is in the same
     *   location where the original undo bits were stored.  We can't
     *   write into the new slot until after the old undo bits are safely
     *   moved out of the way. 
     */
    set_slotptr_propid(get_mod_prop_slot_ptr(old_mod_slots), VM_INVALID_PROP);
}


/* ------------------------------------------------------------------------ */
/*
 *   Look up a property in only the modifiable property slots.  Returns a
 *   pointer to the slot if the property was found among the modifiable
 *   slots, null if not. 
 */
char *CVmObjTads::look_up_mod_prop(vm_prop_id_t prop,
                                   int ret_empty_slot) const
{
    char *slotptr;
    size_t cnt;

    /*
     *   Scan the list of modifiable properties.  This list is in no
     *   particular order, so do a simple linear scan of the list.  
     */
    for (slotptr = get_mod_prop_slot0_ptr(), cnt = get_mod_prop_alloc() ;
         cnt != 0 ; --cnt, inc_prop_slot_ptr(&slotptr))
    {
        /* 
         *   if this slot is empty, return failure, since we've passed the
         *   last slot that's actually in use 
         */
        if (get_slotptr_propid(slotptr) == VM_INVALID_PROP)
        {
            /* 
             *   if we're returning an empty slot pointer, return this
             *   slot; otherwise, indicate return null 
             */
            if (ret_empty_slot)
            {
                /* return the empty slot */
                return slotptr;
            }
            else
            {
                /* 
                 *   they don't want an empty slot - return failure, since
                 *   we didn't find the property among the properties with
                 *   valid values 
                 */
                return 0;
            }
        }
        
        /* if it's a match, return it */
        if (prop == get_slotptr_propid(slotptr))
            return slotptr;
    }
    
    /* 
     *   didn't find the property, and we didn't even find an empty slot -
     *   return null 
     */
    return 0;
}

/* ------------------------------------------------------------------------ */
/*
 *   Look up a property given the property ID, returning a pointer to the
 *   property's data.  The pointer we return points directly into the
 *   property table, so the data holder is in portable format.  We return
 *   null if there is no such property in this object.  This routine does
 *   not look at any superclasses.  
 */
const char *CVmObjTads::look_up_prop(vm_prop_id_t prop) const
{
    ushort hi, lo;
    const char *slotptr;

    /*
     *   Start with the modifiable list, since the modified properties
     *   always take precedence over the original load image properties. 
     */
    if ((slotptr = look_up_mod_prop(prop, FALSE)) != 0)
        return slotptr;

    /*
     *   We didn't find it in the modified list, so look in the load image
     *   list.  This list is sorted in ascending order of property ID, so
     *   we can perform a simple binary search on the array.
     */
    for (lo = 0, hi = get_li_prop_count() ;; )
    {
        ushort cur;

        /* if we've squeezed out everything, we're done */
        if (lo >= hi)
            return 0;

        /* split the difference */
        cur = lo + (hi - lo)/2;

        /* check this slot */
        slotptr = get_li_prop_slot_ptr(cur);
        if (get_slotptr_propid(slotptr) == prop)
            return slotptr;

        /* look above or below, as appropriate */
        if (prop > get_slotptr_propid(slotptr))
        {
            /* we want something higher - look in the upper half */
            lo = (cur == lo ? lo + 1 : cur);
        }
        else
        {
            /* we want something smaller - look in the lower half */
            hi = (cur == hi ? hi - 1 : cur);
        }
    }
}


/* ------------------------------------------------------------------------ */
/*
 *   Get a property.  We first look in this object; if we can't find the
 *   property here, we look for it in one of our superclasses.  
 */
int CVmObjTads::get_prop(VMG_ vm_prop_id_t prop, vm_val_t *val,
                         vm_obj_id_t self, vm_obj_id_t *source_obj,
                         uint *argc)
{
    /* 
     *   try finding the property in our property list or a superclass
     *   properyt list 
     */
    if (get_prop_from_list(vmg_ prop, val, self, source_obj, argc))
        return TRUE;

    /* 
     *   We didn't define the property or inherit it from one of the
     *   objects in our superclass list, so inherit the property from our
     *   base metaclass. 
     */
    return CVmObject::get_prop(vmg_ prop, val, self, source_obj, argc);
}

/* ------------------------------------------------------------------------ */
/*
 *   Inherit a property 
 */
int CVmObjTads::inh_prop(VMG_ vm_prop_id_t prop, vm_val_t *val,
                         vm_obj_id_t self, vm_obj_id_t *source_obj,
                         uint *argc)
{
    /* first, try inheriting the property from a superclass in our list */
    if (inh_prop_from_list(vmg_ prop, val, self, source_obj, argc))
        return TRUE;

    /* 
     *   it's not in any of our superclass's property lists, so inherit
     *   the property from our base metaclass 
     */
    return CVmObject::get_prop(vmg_ prop, val, self, source_obj, argc);
}

/* ------------------------------------------------------------------------ */
/*
 *   Get a property from our property list, or from an inherited property
 *   list 
 */
int CVmObjTads::get_prop_from_list(VMG_ vm_prop_id_t prop, vm_val_t *val,
                                   vm_obj_id_t self, vm_obj_id_t *source_obj,
                                   uint *argc)
{
    const char *slotptr;

    /* first, try to find a slot containing this property */
    slotptr = look_up_prop(prop);
    if (slotptr != 0)
    {
        /* 
         *   we found the property in our own property table - return the
         *   value from our table 
         */
        get_slotptr_val(slotptr, val);
        *source_obj = self;
        return TRUE;
    }

    /*
     *   We didn't find the property in our table.  Try inheriting the
     *   property from a superclass's properyt list.  
     */
    return inh_prop_from_list(vmg_ prop, val, self, source_obj, argc);
}


/* ------------------------------------------------------------------------ */
/*
 *   Inherit a property from an object in our superclass list.
 */
int CVmObjTads::inh_prop_from_list(VMG_ vm_prop_id_t prop, vm_val_t *val,
                                   vm_obj_id_t self, vm_obj_id_t *source_obj,
                                   uint *argc)
{
    size_t sc_cnt;
    size_t i;
    int found;
    vm_obj_id_t inh_src;

    /*
     *   Search each superclass, starting with the first superclass in the
     *   list.
     *   
     *   Even if we find the value in one superclass, continue searching
     *   subsequent superclasses, because we may find a later superclass
     *   that has a common base class with the first superclass, but which
     *   overrides the property from the common base class.  Suppose we
     *   have the following class hierarchy:
     *   
     *   self: A, B
     *.  A: Base { defines P }
     *.  B: C { defines P }
     *.  C: Base { defines Q }
     *.  Base { defines P, Q }
     *   
     *   Suppose we're searching for Q.  We don't have Q in our property
     *   table, so we'll look at superclasses.  We'll start with A; A
     *   doesn't define Q either, so we'll look at A's superclasses.  Base
     *   does define P, so we'll note that we've found Base.Q.  We'll then
     *   continue with our superclasses, moving on to B.  B doesn't define
     *   Q, so we'll look at its superclasses.  C defines Q.  Now, we see
     *   that C.Q overrides Base.Q, because C is derived from Base -- this
     *   means that, even though B is after A in the superclass list,
     *   we'll still inherit Q from C rather than from Base.
     *   
     *   Now, suppose we're searching for P.  We will find that we don't
     *   have P our table, so we'll look at superclasses.  We'll first
     *   look at A, which defines P.  We'll then continue to B.  B also
     *   defines P; however, since B.P does not override A.P (since B is
     *   not derived from B), we'll ignore this and stick with A.P.  
     */
    found = FALSE;
    sc_cnt = get_sc_count();
    for (i = 0 ; i < sc_cnt ; ++i)
    {
        vm_obj_id_t sc;
        CVmObjTads *sc_obj;
        vm_val_t inh_val;
        
        /* get this superclass */
        sc = get_sc(i);
        sc_obj = (CVmObjTads *)vm_objp(vmg_ sc);

        /* if it's not an object of our own type, ignore it */
        if (sc_obj->get_metaclass_reg() != metaclass_reg_)
            continue;

        /* ask this superclass to resolve the property */
        if (sc_obj->get_prop_from_list(vmg_ prop, &inh_val, sc, &inh_src, 0))
        {
            /* 
             *   We found it -- if we've already found one previously,
             *   check to see if this value overrides the one we found
             *   previously, and replace it if so.  If we haven't found
             *   the value until now, use this value. 
             */
            if (!found
                || (vm_objp(vmg_ inh_src)->
                    is_instance_of(vmg_ *source_obj)))
            {
                /* 
                 *   this is the best value we've found so far - remember
                 *   it for return 
                 */
                *source_obj = inh_src;
                *val = inh_val;

                /* note that we've found a value */
                found = TRUE;
            }
        }
    }

    /* if we found a valid object, return true */
    return found;
}

/* ------------------------------------------------------------------------ */
/*
 *   Enumerate properties 
 */
void CVmObjTads::enum_props(VMG_ vm_obj_id_t self,
                            void (*cb)(VMG_ void *ctx, vm_obj_id_t self,
                                       vm_prop_id_t prop,
                                       const vm_val_t *val),
                            void *cbctx)
{
    char *slotptr;
    size_t cnt;
    size_t i;
    vm_val_t val;
    size_t sc_cnt;

    /* first, go through the modified properties */
    for (slotptr = get_mod_prop_slot0_ptr(), cnt = get_mod_prop_alloc() ;
         cnt != 0 ; --cnt, inc_prop_slot_ptr(&slotptr))
    {
        /* 
         *   if this slot is empty, we've reached the end of the modified
         *   properties 
         */
        if (get_slotptr_propid(slotptr) == VM_INVALID_PROP)
            break;

        /* get this property's value */
        get_slotptr_val(slotptr, &val);

        /* invoke the callback for this property */
        (*cb)(vmg_ cbctx, self, get_slotptr_propid(slotptr), &val);
    }

    /* 
     *   now enumerate the original properties that don't appear in the
     *   modified list 
     */
    for (i = 0, cnt = get_li_prop_count() ; i < cnt ; ++i)
    {
        const char *slotptr;
        
        /* get this slot */
        slotptr = get_li_prop_slot_ptr(i);

        /* if it isn't in the modified properties, call the callback */
        if (look_up_mod_prop(get_slotptr_propid(slotptr), FALSE) == 0)
        {
            /* it's not in the modified properties - get the value */
            get_slotptr_val(slotptr, &val);

            /* invoke the callback for this property */
            (*cb)(vmg_ cbctx, self, get_slotptr_propid(slotptr), &val);
        }
    }

    /*
     *   enumerate properties in each superclass 
     */
    sc_cnt = get_sc_count();
    for (i = 0 ; i < sc_cnt ; ++i)
    {
        vm_obj_id_t sc;

        /* get this superclass */
        sc = get_sc(i);

        /* enumerate its properties */
        vm_objp(vmg_ sc)->enum_props(vmg_ sc, cb, cbctx);
    }
}


/* ------------------------------------------------------------------------ */
/*
 *   Determine if I'm an instance of the given object 
 */
int CVmObjTads::is_instance_of(VMG_ vm_obj_id_t obj)
{
    size_t i;
    size_t sc_cnt;

    /* 
     *   for each superclass, determine whether the superclass matches the
     *   given object, or if the superclass derives from the given object 
     */
    sc_cnt = get_sc_count();
    for (i = 0 ; i < sc_cnt ; ++i)
    {
        vm_obj_id_t sc;

        /* get this superclass */
        sc = get_sc(i);

        /* 
         *   if it matches the object we're looking for, this object
         *   derives directly from the given object 
         */
        if (sc == obj)
            return TRUE;

        /* 
         *   if this superclass derives from the given object, this object
         *   derives indirectly from the given object 
         */
        if (vm_objp(vmg_ sc)->is_instance_of(vmg_ obj))
            return TRUE;
    }

    /* 
     *   None of our superclasses match the given object, and none of the
     *   superclasses derive from the given object, so we must not derive
     *   from the given object.  Our last recourse is to determine if the
     *   object represents our metaclass; inherit the default handling to
     *   make this check.  
     */
    return CVmObject::is_instance_of(vmg_ obj);
}

/* ------------------------------------------------------------------------ */
/*
 *   Apply undo 
 */
void CVmObjTads::apply_undo(VMG_ CVmUndoRecord *rec)
{
    char *slotptr;
    
    /* 
     *   Find the property slot among the modified properties - undo
     *   obviously only applies to the modified properties.  If we can't
     *   find the property, there must be some problem; there's nothing we
     *   can do about it, so just ignore it 
     */
    slotptr = look_up_mod_prop(rec->id.prop, FALSE);
    if (slotptr == 0)
    {
        assert(FALSE);
        return;
    }

    /* 
     *   if the record indicates that the property was originally empty,
     *   clear the slot entirely; otherwise, restore the value from the
     *   record 
     */
    if (rec->oldval.typ == VM_EMPTY)
    {
        /* 
         *   It didn't previously exist - delete this slot.  Since undo is
         *   always applied in reverse order, and we always add new
         *   properties at the end of the modified property list, deleting
         *   this property simply means setting its ID to 'invalid' --
         *   there can be no properties after this one, because it is not
         *   possible for other operations to have been performed (and
         *   still be in effect) after the operation that created this
         *   property entry.  
         */
        set_slotptr_propid(slotptr, VM_INVALID_PROP);
    }
    else
    {
        /* restore the original value */
        set_prop_slotptr(slotptr, rec->id.prop, &rec->oldval);
    }
}


/* ------------------------------------------------------------------------ */
/*
 *   Mark as referenced all of the objects to which we refer 
 */
void CVmObjTads::mark_refs(VMG_ uint state)
{
    size_t slot;
    size_t slotcnt;
    char *slotptr;
    
    /* 
     *   Go through all of our property slots and mark each object value.
     *   Note that we only need to worry about the modified properties;
     *   everything referenced in the load image list is necessarily part
     *   of the root set, or it couldn't have been in the load image, so
     *   we don't need to bother marking any of those objects, since they
     *   can never be deleted by virtue of being in the root set.  
     */
    slotcnt = get_mod_prop_alloc();
    for (slot = 0, slotptr = get_mod_prop_slot_ptr(0) ; slot < slotcnt ;
         ++slot, inc_prop_slot_ptr(&slotptr))
    {
        vm_val_t val;

        /* if the slot isn't in use, we're finished */
        if (get_slotptr_propid(slotptr) == VM_INVALID_PROP)
            break;
            
        /* get the value from the slot */
        get_slotptr_val(slotptr, &val);

        /* if it's an object value, mark the object as referenced */
        if (val.typ == VM_OBJ && val.val.obj != VM_INVALID_OBJ)
            G_obj_table->mark_all_refs(val.val.obj, state);
    }
}


/* ------------------------------------------------------------------------ */
/*
 *   Mark a reference in an undo record 
 */
void CVmObjTads::mark_undo_ref(VMG_ CVmUndoRecord *undo)
{
    /* if the undo record refers to an object, mark the object */
    if (undo->oldval.typ == VM_OBJ)
        G_obj_table->mark_all_refs(undo->oldval.val.obj, VMOBJ_REACHABLE);
}

/* ------------------------------------------------------------------------ */
/*
 *   Determine if the object has been changed since it was loaded from the
 *   image file.  If the object has no properties stored in the modified
 *   properties table, it is in exactly the same state as is stored in the
 *   image file.  
 */
int CVmObjTads::is_changed_since_load() const
{
    /* 
     *   if we don't have a modified properties table, or the first entry
     *   is empty, the object hasn't been modified; if we have any
     *   non-empty entries, the object has been modified 
     */
    return (get_mod_prop_alloc() != 0
            && (get_slotptr_propid(get_mod_prop_slot0_ptr())
                != VM_INVALID_PROP));
}

/* ------------------------------------------------------------------------ */
/*
 *   Save the object's state to a file.  We only need to save the modified
 *   property list, because the load image list never changes.  
 */
void CVmObjTads::save_to_file(VMG_ CVmFile *fp)
{
    uint cnt;
    char *slot;

    /* count the number of properties that have actually been modified */
    for (cnt = 0, slot = get_mod_prop_slot0_ptr() ;
         cnt < get_mod_prop_alloc() ; ++cnt, inc_prop_slot_ptr(&slot))
    {
        /* 
         *   if this slot is empty, stop here - once we find the first
         *   empty slot, all remaining slots must be empty 
         */
        if (get_slotptr_propid(slot) == VM_INVALID_PROP)
            break;
    }

    /* write the number of modified properties */
    fp->write_int2(cnt);

    /* 
     *   if we didn't come from the image file, save the superclass list -
     *   we only need to do this for objects created dynamically, because
     *   the superclass list for an image file object cannot be changed
     *   after loading the object from the image file 
     */
    if (!is_from_image())
    {
        /* write the number of superclasses */
        fp->write_int2(get_sc_count());

        /* write the superclass list */
        fp->write_bytes(get_sc_slot0_ptr(), get_sc_count() * VMB_OBJECT_ID);
    }

    /* write the modified property data */
    fp->write_bytes(get_mod_prop_slot0_ptr(), calc_save_size(cnt));
}

/* ------------------------------------------------------------------------ */
/*
 *   Restore the object from a file 
 */
void CVmObjTads::restore_from_file(VMG_ vm_obj_id_t,
                                   CVmFile *fp, CVmObject *obj)
{
    ushort mod_count;
    
    /* read number of modified properties */
    mod_count = fp->read_uint2();

    /* 
     *   If we don't have an extension yet, allocate one.  The only way we
     *   won't have an extension is if we weren't loaded from the image
     *   file, since we always create the extension upon construction when
     *   loading from an image file.  
     */
    if (ext_ == 0)
    {
        ushort sc_cnt;

        /* 
         *   read the superclass count - we need to know this, in addition
         *   to the modified property count (which we've already read), to
         *   know how much space to allocate 
         */
        sc_cnt = fp->read_uint2();

        /* allocate our extension */
        alloc_mem(vmg_ sc_cnt, mod_count, FALSE);

        /* 
         *   read our superclass list - since we didn't come from the
         *   image file, we need to get our superclass list from the saved
         *   state file
         */
        fp->read_bytes((char *)get_sc_slot0_ptr(), sc_cnt * VMB_OBJECT_ID);
    }
    else
    {
        /* 
         *   We already have an extension, so we must have come from the
         *   image file.  Make sure we have enough memory to hold this
         *   many properties.
         */
        expand_prop_table(vmg_ mod_count, obj);
    }

    /* read the persistent data */
    fp->read_bytes(get_mod_prop_slot_ptr(0),
                   calc_save_size(mod_count));

    /* 
     *   clear the next slot after the slots we read, if we indeed have
     *   more slots after the ones we read
     */
    if (get_mod_prop_alloc() > mod_count)
        set_slotptr_propid(get_mod_prop_slot_ptr(mod_count), VM_INVALID_PROP);

    /* forget all of our undo bits */
    clear_undo_bits();
}

/* ------------------------------------------------------------------------ */
/*
 *   Load the object from an image file
 */
void CVmObjTads::load_from_image(VMG_ vm_obj_id_t,
                                 const char *ptr, size_t siz)
{
    /* if we already have memory allocated, free it */
    if (ext_ != 0)
    {
        G_mem->get_var_heap()->free_mem(ext_);
        ext_ = 0;
    }

    /* allocate memory */
    alloc_mem(vmg_ 0, VMTOBJ_PROP_INIT, TRUE);

    /* set the image data pointer */
    get_hdr()->image_data = ptr;
}

/* ------------------------------------------------------------------------ */
/*
 *   Reset to image file state.  Discards all modified properties, so that
 *   we have only the image file properties.
 */
void CVmObjTads::reset_to_image(VMG0_)
{
    /* if we have any modified property slots, clear the first one */
    if (get_mod_prop_alloc() != 0)
        set_slotptr_propid(get_mod_prop_slot0_ptr(), VM_INVALID_PROP);

    /* forget all of our undo bits */
    clear_undo_bits();
}

/* ------------------------------------------------------------------------ */
/*
 *   Allocate memory for a given configuration.  
 */
void CVmObjTads::alloc_mem(VMG_ ushort sc_count,
                           ushort mod_prop_count, int from_image)
{
    size_t siz;

    /* calculate the size required for this configuration */
    siz = calc_alloc_size(sc_count, mod_prop_count, from_image);
    ext_ = (char *)G_mem->get_var_heap()->alloc_mem(siz, this);

    /* 
     *   if we're not loaded from the image file, set the image data
     *   pointer to refer to our own extension, following the non-image
     *   data 
     */
    if (!from_image)
    {
        /* 
         *   set the image data pointer, calculating the size that we'd
         *   have for our ordinary non-image data only 
         */
        get_hdr()->image_data =
            ext_ + calc_alloc_size(0, mod_prop_count, TRUE);

        /* set the superclass count in our fake image data */
        set_sc_count(sc_count);

        /* we have no fixed properties, since we're not from an image file */
        set_li_prop_count(0);

        /* set the object flags */
        set_obj_flags(0);
    }

    /* set the modifiable property count */
    set_mod_prop_alloc(mod_prop_count);

    /* set the first modifiable property to empty */
    set_slotptr_propid(get_mod_prop_slot_ptr(0), VM_INVALID_PROP);
}

